{-# OPTIONS -fglasgow-exts #-}

module Syntax.ConcreteToInternal
    ( programToInternal
    , termToInternal
    , termToInternalM
    , Env(..), emptyEnv
    , runConvM
    , programToInternalM
    ) where

import Control.Monad.State
import Control.Monad.Writer
import Control.Monad.Reader
import Data.Set ( Set, addToSet, delFromSet, minusSet, unitSet
		, elementOf, unionManySets, union, emptySet, mkSet
		)
import Data.List (intersperse)

import qualified Syntax.Concrete as C
import qualified Syntax.Internal as I

import Syntax.Name
import Syntax.Position
import Syntax.Pretty	-- TODO: add pretty printing to error messages

import Utils.Monads
import Utils.Map

-- Massaging the concrete declarations ------------------------------------

data Definition	= Axiom Name C.Type
		| Def Name (Maybe C.Type) C.Term

addFixity x f =
    do	inMap <- isInMap x <$> get
	if inMap
	    then error $ "Parse error: Multiple fixities for " ++ x
	    else modify $ updateMap x f

-- | Rearrange a list of concrete declarations into fixity declarations
--   and concrete function definitions. This function could potentially
--   do strongly connected components and reordering of definitions as well.
definitions :: [C.Decl] -> (Map Name C.Fixity, [Definition])
definitions ds = runWriter $ execStateT (defs ds) emptyMap
    where
	defs ds =
	    case ds of

		[] -> return ()

		sig@(C.TypeSig n _) : ds' ->
		    case span (defOf n) ds' of
			(clauses, ds'')	->
			    do	mkFun (Just sig) clauses
				defs ds''

		C.Infix a p xs : ds'	->
		    do	mapM_ (\x -> addFixity x (a,p)) xs
			defs ds'

		C.FunClause n _ : ds' ->
		    case span (defOf n) ds of
			(clauses, ds'') ->
			    do	mkFun Nothing clauses
				defs ds''

	defOf n (C.FunClause n' _) = n == n'
	defOf _ _ = False

	-- Precondition for @mkFun t cs@: @t == Nothing ==> not (null cs)@
	mkFun Nothing [clause@(C.FunClause n _)] =
		tell [Def n Nothing $ mkClause clause]
	mkFun (Just (C.TypeSig n t)) [] =
		tell [Axiom n t]
	mkFun (Just (C.TypeSig n t)) [clause] =
		tell [Def n (Just t) $ mkClause clause]
	
	mkClause (C.FunClause _ t) = t

-- Conversion to internal syntax ------------------------------------------

-- Thierry:
-- the constants are represented abstractly directly as a name
-- and a pointer to their definitions (like McCarthy's notion of
-- labels). This looks simpler than having "rho" and "defs" as
-- arguments for the function valueOf, and can be done when
-- translating Concrete syntax -> Internal syntax (we check the
-- scope of identifiers at this stage)

type ConvM = Reader Env
data Env    = Env   { eDefs	    :: Set Name
		    , eFixities	    :: Map Name C.Fixity
		    }

emptyEnv = Env emptySet emptyMap

addDef x	= local $ \e -> e { eDefs	    = eDefs e `addToSet` x }
bindName x	= local $ \e ->
		    e   { eDefs		= eDefs e `delFromSet` x
			}
bindNames xs	= local $ \e ->
		    e   { eDefs		= eDefs e `minusSet` mkSet xs
			}

getFixity x =
    maybe C.defaultFixity id <$> (lookupMap x <$> (eFixities <$> ask))

-- | The continuation gets the list of clashing names.
addFixities :: Map Name C.Fixity -> ([Name] -> ConvM a) -> ConvM a
addFixities fxs' f =
    do	fxs <- eFixities <$> ask
	let clashes = keysMap (intersectMap fxs fxs')
	let fxs''   = plusMap fxs fxs'
	local (\e -> e { eFixities = fxs'' }) (f clashes)

mkName x =
    do	e <- ask
	return $
	    if x `elementOf` eDefs e then
		I.Def x ()
	    else
		I.Var x ()

runConvM = flip runReader emptyEnv

programToInternal :: C.Program -> I.Program
programToInternal p =
    fst $ runConvM $ (programToInternalM p $ return ())

programToInternalM :: C.Program -> ConvM a -> ConvM (I.Program, a)
programToInternalM p k =
    addFixities fixs $ \clashes ->
	if null clashes
	then definitionsToInternalM ds k
	else error $ "Parse error: Repeated fixity declarations for " ++ concat (intersperse ", " clashes)
    where
	(fixs, ds) = definitions p

definitionsToInternalM :: [Definition] -> ConvM a -> ConvM (I.Program, a)
definitionsToInternalM = c2i
    where
	c2i [] k = (,) [] <$> k

	c2i (Def name mt c : p) k =
	    do	d <- I.NormalDef name
			<$> mapMaybeM termToInternalM mt
			<*> addDef name (termToInternalM c)
		(p, x) <- addDef name (c2i p k)
		return (d:p, x)
	c2i (Axiom name t : p) k =
	    do	d <- I.NewAxiom name <$> termToInternalM t
		(p, x) <- addDef name (c2i p k)
		return (d:p, x)

termToInternal :: C.Term -> I.Term
termToInternal t = runReader (termToInternalM t) emptyEnv

-- Note: Maybe we should handle variable shadowing here?
termToInternalM :: C.Term -> ConvM I.Term
termToInternalM = c2i
    where
	c2i (C.App e1 e2)	    = I.App <$> c2i e1 <*> c2i e2
	c2i (C.Name x)	    = mkName x
	c2i (C.Lam x e)	    = I.Lam x	<$> bindName x (c2i e)
						<*> return ()
	c2i (C.Star (C.Univ n))   = return $ I.Star n
	c2i C.Prop		    = return I.Prop
	c2i (C.Pi x a b)	    = I.Pi x	<$> c2i a
						<*> bindName x (c2i b)
						<*> return ()
        c2i (C.CallPlugin n es)   = I.CallPlugin n
					    <$> mapM c2i es
					    <*> return ()
	c2i (C.Let p e)	    =
	    addFixities fixs $ \clashes ->
		if null clashes
		then letToInternal ds e
		else error $ "Parse error: Repeated fixity declarations for " ++ concat (intersperse ", " clashes)
	    where
		(fixs, ds) = definitions p

	c2i (C.Paren e)	    = c2i e

	c2i e@(C.InfixApp _ _ _)   = c2i =<< (unInfix <$> rotate e)

	unInfix (C.InfixApp e1 op e2) =
	    C.Name op `C.App` unInfix e1 `C.App` unInfix e2
	unInfix e = e

	rotate e@(C.InfixApp e1 op e2) =
	    do	e1' <- rotate e1
		case e1' of
		    C.InfixApp e1a op1 e1b ->
			do  fx	<- getFixity op
			    fx1	<- getFixity op1
			    if shouldRotate op1 op fx1 fx
				then do	e3 <- rotate (C.InfixApp e1b op e2)
					return $ C.InfixApp e1a op1 e3
				else return $ C.InfixApp e1' op e2
		    _ -> return e
	
	rotate e = return e

	letToInternal :: [Definition] -> C.Term -> ConvM I.Term
	letToInternal [] e = I.Let [] <$> c2i e
	letToInternal (Def x mt e0 : defs) e =
		do  I.Let defs' e' <- bindName x (letToInternal defs e)
		    def' <- I.LocalDef x
				<$> mapMaybeM c2i mt
				<*> c2i e0
				<*> return ()
		    return $ I.Let (def':defs') e'
	letToInternal (Axiom x _ : _) _ =
		error $ "Parse error:\n You cannot define local axioms (" ++ x ++ ")"

shouldRotate op1 op2 (a1,p1) (a2,p2) =
    case (a1,a2) of
	(C.RightAssoc, C.RightAssoc)	-> p2 >= p1
	(C.LeftAssoc,	C.LeftAssoc)	-> p2 > p1
	_   | p1 == p2			->
		error $ "Parse error"
			++ ":\n  Cannot mix infix applications of (" ++ op1
			++ ") and (" ++ op2 ++ ")"
	    | otherwise			-> p2 > p1

