{-# OPTIONS -fglasgow-exts #-}

module Syntax.ParseMonad where

import Control.Monad.State
import Control.Monad.Error

import Syntax.Position
import Utils.Monads

-- Stuff that Alex needs --------------------------------------------------

type AlexInput = (Posn, String, Char)

alexInputPrevChar :: AlexInput -> Char
alexInputPrevChar (_,_,c) = c

alexGetChar :: AlexInput -> Maybe (Char, AlexInput)
alexGetChar (_, [], _)	= Nothing
alexGetChar (p, c:s, _)	= p' `seq` Just (c, (p', s, c))
    where p' = movePos p c

-- The monad --------------------------------------------------------------

type LexState = Int

newtype Parser a = P { unP :: ParseState -> ParseResult a }

data ParseState = PState
    { parsePos	    :: !Posn		-- position at current input location
    , parseLastPos  :: !Posn		-- position of last token
    , parseInp	    :: String		-- the current input
    , parsePrevChar :: !Char		-- the character before the input
    , parsePrevToken:: String		-- the previous token
    , parseLayout   :: [LayoutContext]	-- the current layout context
    , parseLexState :: [LexState]	-- the state of the lexer
    }
    deriving Show

data ParseResult a  = ParseOk ParseState a
		    | ParseFailed Posn String String
    deriving Show

instance Monad (Parser) where
    return x	= P $ \s -> ParseOk s x
    P m >>= f	= P $ \s -> case m s of
				ParseFailed p t e   -> ParseFailed p t e
				ParseOk s' x	    -> unP (f x) s'
    fail e	= P $ \s -> ParseFailed (parseLastPos s) (parsePrevToken s) e

instance MonadError String Parser where
    throwError = fail
    P m `catchError` h	= P $ \s -> case m s of
					ParseFailed p t e   -> unP (h e) s
					m'		    -> m'

instance MonadState ParseState Parser where
    get = P $ \s -> ParseOk s s
    put s = P $ \_ -> ParseOk s ()

showParseFailure p why what =
	what ++ " at " ++ show p ++ ": Unexpected " ++ showWhy why ++ "."
    where
	showWhy ""  = "end of file"
	showWhy t   = "'" ++ t ++ "'"

initState :: FilePath -> String -> LexState -> ParseState
initState file inp st =
	PState  { parsePos	    = p
		, parseLastPos	    = p
		, parseInp	    = inp
		, parsePrevChar	    = '\n'
		, parsePrevToken    = ""
		, parseLexState	    = [st]
		, parseLayout	    = []
		}
    where
	p = startPos file

parse' :: LexState -> Parser a -> String -> ParseResult a
parse' st p input = unP p (initState "" input st)

parseFile' :: LexState -> Parser a -> FilePath -> IO (ParseResult a)
parseFile' st p file =
    do	input <- liftIO $ readFile file
	return $ unP p (initState file input st)

getLexInput :: Parser AlexInput
getLexInput = getInp <$> get
    where
	getInp s = (parsePos s, parseInp s, parsePrevChar s)

setLexInput :: AlexInput -> Parser ()
setLexInput (p,i,c) = modify upd
    where
	upd s = s { parsePos = p, parseInp = i, parsePrevChar = c }

setLastPos :: Posn -> Parser ()
setLastPos p = modify $ \s -> s { parseLastPos = p }

setPrevToken :: String -> Parser ()
setPrevToken t = modify $ \s -> s { parsePrevToken = t }

getParsePos :: Parser Posn
getParsePos = get >>= return . parseLastPos

getParseRange :: Parser Range
getParseRange =
    do	s <- get
	return $ Range (parseLastPos s) (parsePos s)

parseError :: String -> Parser a
parseError message = fail message

parseErrorAt :: Posn -> String -> Parser a
parseErrorAt p msg = setLastPos p >> fail msg

displayChar :: Char -> String
displayChar = init . tail . show

getLexState :: Parser [LexState]
getLexState = parseLexState <$> get

setLexState :: [LexState] -> Parser ()
setLexState ls =
    do	s <- get
	put $ s { parseLexState = ls }

pushLexState :: LexState -> Parser ()
pushLexState l = do s <- getLexState
		    setLexState (l:s)

popLexState :: Parser ()
popLexState = do    _:ls <- getLexState
		    setLexState ls

-- -----------------------------------------------------------------------------
-- Useful token actions

-- | LexAction result = previous_input ->
--			current_input ->
--			length_of_token -> Parser result
type LexAction r = AlexInput -> AlexInput -> Int -> Parser r

constant :: t -> LexAction t
constant t  = scanTok $ \_ -> return t

tok :: (String -> r) -> LexAction r
tok t = scanTok $ \s -> return $ t s

withPos :: (Posn -> String -> r) -> LexAction r
withPos f = scanTok $ \s ->
			do  p <- getParsePos
			    return $ f p s

withPos_ :: (Posn -> r) -> LexAction r
withPos_ f = withPos (\p _ -> f p)

withRange :: (Range -> String -> r) -> LexAction r
withRange f = scanTok $ \s ->
			do  r <- getParseRange
			    return $ f r s

withRange_ :: (Range -> r) -> LexAction r
withRange_ f = withRange (\r _ -> f r)

-- Scan functions ---------------------------------------------------------

scanEOF (p,i,_) eof =
    do  setLastPos p
	setPrevToken "EOF"
	return eof

scanError (p,i,_) =
    do  setLastPos p
	setPrevToken $ case i of
			c:_ -> displayChar c
			_	-> "EOF"
	parseError "Lexical error"

scanSkip inp r = setLexInput inp >> r

scanTok action (p,i,_) inp' len =
    do  setLexInput inp'
	setPrevToken t
	setLastPos p
	action t
    where
	t = take len i

-- Layout -----------------------------------------------------------------

data LayoutContext  = NoLayout
		    | Layout Int
    deriving Show

getContext :: Parser [LayoutContext]
getContext = parseLayout <$> get

setContext :: [LayoutContext] -> Parser ()
setContext ctx =
    do	s <- get
	put $ s { parseLayout = ctx }

popContext :: Parser ()
popContext =
    do	_:ctx <- getContext
	setContext ctx

pushContext :: LayoutContext -> Parser ()
pushContext l =
    do	ctx <- getContext
	setContext (l : ctx)

-- | Only at the beginning of the file
pushCurrentContext :: Parser ()
pushCurrentContext =
    do	p <- getParsePos
	pushContext (Layout (posCol p))

-- | Compute the relative position of a location to the
--   current layout context.
getOffside :: Posn -> Parser Ordering
getOffside loc =
    do	ctx <- getContext
	return $ case ctx of
	    Layout n : _    -> compare (posCol loc) n
	    _		    -> GT

