{-# OPTIONS -fglasgow-exts #-}

module Syntax.ConcreteToInternal
    ( programToInternal
    , termToInternal
    , termToInternalM
    , ScopeEnv(..), emptyScopeEnv
    , runConvM, runConvM'
    , programToInternalM
    , addDefinedName, addVarNames
    ) where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.State
import Control.Monad.Reader
import Control.Monad.Writer
import Control.Monad.Error

import Data.Set as Set
		( Set, insert, delete, difference, singleton
		, member, unions, union, empty, fromList, toList
		)
import qualified Data.List as List

import qualified Syntax.Concrete as C
import qualified Syntax.Internal as I

import Syntax.Name
import Syntax.Position
import Syntax.Pretty	-- TODO: add pretty printing to error messages
import Syntax.Scope

import Utils.Names
import Utils.Monads
import Utils.Continuations
import Utils.Functions
import qualified Utils.Map as Map

-- Massaging the concrete declarations ------------------------------------

data Definition	= Axiom Range C.Name C.Telescope C.Type
		| Synonym Range C.Name C.Term
		| Def Range DefType DefValue
		| Rewrite Range [C.Term]
		| Mutual Range [(DefType,DefValue)]
		| Abstract Range [(DefType,DefValue)]

data DefType = DataType Bool Range C.Name C.Telescope C.Telescope C.Type
	     | FunType Range C.Name I.Arity C.Telescope C.Type
	     | SigType Range C.Name C.Telescope C.Telescope C.Type [C.Name]

data DefValue = DataVal [C.Constructor]
	      | FunVal [Clause]
	      | SigVal [C.Constructor]

data Clause	= Clause Range C.Name C.Clause

instance HasRange Definition where
    getRange (Axiom r _ _ _)	    = r
    getRange (Def r _ _)	    = r
    getRange (Rewrite r _)	    = r
    getRange (Mutual r _)	    = r
    getRange (Abstract r _)	    = r
    getRange (Synonym r _ _)	    = r

addFixity r x f =
    do	inMap <- Map.member x <$> get
	if inMap
	    then fail $ "Parse error at " ++ show r ++ ": Multiple fixities for " ++ x
	    else modify $ Map.insert x f

-- | Rearrange a list of concrete declarations into fixity declarations
--   and concrete function definitions. This function could potentially
--   do strongly connected components and reordering of definitions as well.
definitions :: [C.Decl] -> (Map.Map String C.Fixity, [Definition])
definitions ds = runWriter $ execStateT (defs ds) Map.empty
    where
	defs ds =
	    case ds of

		[] -> return ()

		sig@(C.TypeSig _ n tel _) : ds' ->
		    case span (defOf n) ds' of
			(clauses, ds'')	->
			    do	mkFun (Just sig) clauses
				defs ds''

		C.Rewrite r e : ds' ->
		    do	tell [ Rewrite r e ]
			defs ds'

		C.Mutual r ds : ds' ->
		    do	(_,ds0) <- censor (const [])
				   $ listen $ defs ds
			ds0' <- mapM unDef ds0
			tell [Mutual r ds0']
			defs ds'

		C.Abstract r ds : ds' ->
		    do	(_,ds0) <- censor (const [])
				   $ listen $ defs ds
			ds0' <- mapM unDef ds0
			tell [Abstract r ds0']
			defs ds'

		C.Infix r a p xs : ds'	->
		    do	mapM_ (\ (C.Name _ x) -> addFixity r x (a,p)) xs
			defs ds'

		C.Data sized r x htel tel t cs : ds' ->
		    do	tell [ Def r (DataType sized r x htel tel t)
				     (DataVal cs)
			     ]
			defs ds'

		C.Signature r x htel tel t ps cs : ds' ->
		    do	tell [ Def r (SigType r x htel tel t ps)
				     (SigVal cs)
			     ]
			defs ds'

		C.FunClause _ n _ : ds' ->
		    case span (defOf n) ds of
			(clauses, ds'') ->
			    do	mkFun Nothing clauses
				defs ds''

	defOf n (C.FunClause _ n' _) = n == n'
	defOf _ _ = False

	unDef (Def _ t v) = return (t,v)
	unDef d	= fail $ show (getRange d) ++ ": Only function definitions and data/signature declarations are allowed in 'mutual' or 'abstract'"

	-- Precondition for @mkFun t cs@: @t == Nothing ==> not (null cs)@
	mkFun Nothing [C.FunClause r n (C.Clause _ [] [] (C.NoWith e))] =
	    tell [Synonym (fuseRange r e) n e]
	mkFun (Just (C.TypeSig r n tel t)) [] =
	    tell [ Axiom r n tel t ]
	mkFun (Just (C.TypeSig r n tel t)) clauses =
	    do	a <- getArity r' n clauses
		tell [ Def r' (FunType r' n a tel t)
			      (FunVal $ map (mkClause tel) clauses)
		     ]
	    where
		r' = fuseRange r (last clauses)
	mkFun t ds = fail $ "Bad function definition at " ++ show (fuseRange t ds)

	mkClause tel (C.FunClause r x c) =
		Clause r x $ padHidden hps c
	    where
		hps = map (var . fst) tel
		var x = C.WildP (getRange x)
	mkClause _ _ = error "impossible: Syntax.ConcreteToInternal.mkClause"

	padHidden hs' (C.Clause r hs ps w) =
		C.Clause r (hs ++ hsp) ps (padHiddenWith hs' w)
	    where
		hsp = drop (length hs) hs'

	padHiddenWith hs' (C.NoWith e) = C.NoWith e
	padHiddenWith hs' (C.With es cs) = C.With es $ map (padHidden hs') cs

	getArity r x cs = foldr1 op $ map (return . clauseArity) cs
	    where
		op cn cm =
		    do	n <- cn
			m <- cm
			if n == m then return n
				  else fail $ "Parse error at " ++ show r ++ ": Equations give different arities for " ++ show x

	clauseArity (C.FunClause _ _ (C.Clause _ _ ps _)) = length ps
	clauseArity _ = error "impossible: Syntax.ConcreteToInternal.clauseArity"

-- Conversion to internal syntax ------------------------------------------

-- Thierry:
-- the constants are represented abstractly directly as a name
-- and a pointer to their definitions (like McCarthy's notion of
-- labels). This looks simpler than having "rho" and "defs" as
-- arguments for the function valueOf, and can be done when
-- translating Concrete syntax -> Internal syntax (we check the
-- scope of identifiers at this stage)

type ConvM = StateT UsedNames (ReaderT ScopeEnv (Either String))

lookupName x =
    do	e <- ask
	return $ Map.lookup x (eNames e)

getName :: C.Name -> ConvM I.Term
getName name@(C.Name r x) =
    do	mx <- lookupName x
	case mx of
	    Nothing -> I.Var r <$> freshName VarName name
		    --fail $ show r ++ ": Undefined variable " ++ show x
	    Just x' -> return $ bnameToIName $ setRange r x'
    where
	setRange r (BName k x) = BName k $ x { nameRange = r }
getName C.DontCare  = error $ "impossible: Syntax.ConcreteToInternal.getName DontCare"

getNameOfKind :: Kind -> C.Name -> ConvM I.Name
getNameOfKind k (C.Name r x) =
    do	mx <- lookupName x
	case mx of
	    Nothing -> fail $ show r ++ ": Undefined variable " ++ show x
	    Just (BName k' x')
		| k == k'   -> return $ x' { nameRange = r }
		| otherwise -> fail $ "expected " ++ x ++ " to be a " ++ show k ++ ", not a " ++ show k'
getNameOfKind _ C.DontCare = return I.DontCare

bnameToIName (BName k x) =
    case k of
	DefName	 -> I.Def r x
	DataName -> I.Data r x
	ConName	 -> I.Con r x
	VarName	 -> I.Var r x
	SigName	 -> I.Sig r x
    where
	r = getRange x

freshName' :: (BoundName -> BoundName -> ConvM BoundName) ->
	      Kind -> C.Name -> ConvM BoundName
freshName' _ k C.DontCare = return $ BName k dontCare
freshName' clash k (C.Name r x) =
    do	new <- setRange r . BName k <$> fresh x
	mx <- lookupName x
	case mx of
	    Nothing	-> return new
	    Just old	->
		do  real <- clash new old
		    return $ setRange r real
    where
	setRange r (BName k x) = BName k $ x { nameRange = r }

freshName k x	 = actualName <$> freshName' normalClash k x
freshPatternName = freshName' patternClash VarName
freshFieldName x = actualName <$> freshName' fieldClash VarName x

patternClash x y = return y
fieldClash x y	 = return x
normalClash x y	 = return x --fail $ show rx ++ ": Duplicate definition of " ++ show x ++ ", previous binding at " ++ show ry
    where
	rx = getRange x
	ry = getRange y

bindName :: Kind -> I.Name -> ConvM a -> ConvM a
bindName k x = local $ \e -> e { eNames = addName (BName k x) $ eNames e }

addSignature :: I.Name -> Set.Set I.Name -> ConvM a -> ConvM a
addSignature s fs =
    local $ \env -> env { eSignatures = Map.insert s fs $ eSignatures env }

getSignature :: I.Name -> ConvM (Set.Set I.Name)
getSignature s =
    do	fs <- Map.lookup s . eSignatures <$> ask
	case fs of
	    Nothing -> fail $ show s ++ " should be a signature name"
	    Just fs -> return fs

bindVariables = bindNames VarName
bindNames k xs = foldr (.) id $ map (bindName k) $ Set.toList xs

bindTelescope tel = bindVariables (Set.fromList $ map fst tel)

addConstrs :: [Name] -> ConvM a -> ConvM a
addConstrs cs k = foldr ($) k $ map (bindName ConName) cs

getFixity (C.Name _ x) =
    maybe C.defaultFixity id <$> (Map.lookup x <$> (eFixities <$> ask))
getFixity _ = error "impossible: Syntax.ConcreteToInternal.getFixity DontCare"

-- | The continuation gets the list of clashing names.
addFixities :: Map.Map String C.Fixity -> ([String] -> ConvM a) -> ConvM a
addFixities fxs' f =
    do	fxs <- eFixities <$> ask
	let clashes = Map.keys (Map.intersection fxs fxs')
	let fxs''   = Map.union fxs fxs'
	local (\e -> e { eFixities = fxs'' }) (f clashes)

checkForFieldNameClashes :: Range -> Set.Set I.Name -> ConvM ()
checkForFieldNameClashes r fs =
    case clash of
	[]	-> return ()
	xs:_	-> fail $ show r ++ ": Conflicting fields " ++ show (head xs) ++ " " ++ show (map getRange xs)
    where
	clash = filter ((>1) . length)
		$ List.groupBy ((==) **.* nameString)
		$ List.sortBy (compare **.* nameString)
		$ Set.toList fs

runConvM = runConvM' noNamesUsed emptyScopeEnv
runConvM' used scope = flip runReaderT scope . flip evalStateT used

programToInternal :: C.Program -> Either String (I.Program, (UsedNames, ScopeEnv))
programToInternal ps =
    runConvM $ (programToInternalM ps $ (,) <$> get <*> ask)

programToInternalM :: C.Program -> ConvM a -> ConvM (I.Program, a)
programToInternalM p k =
    addFixities fixs $ \clashes ->
	if null clashes
	then definitionsToInternalM ds k
	else fail $ "Parse error: Repeated fixity declarations for " ++ concat (List.intersperse ", " clashes)
    where
	(fixs, ds) = definitions p

definitionsToInternalM :: [Definition] -> ConvM a -> ConvM (I.Program, a)
definitionsToInternalM = c2i
    where
	c2i [] k = (,) [] <$> k

	c2i (Synonym r x e : p) k =
	    do	x' <- freshName DefName x
		d <- I.Synonym r x' <$> termToInternalM e
		(p,x) <- bindName DefName x' (c2i p k)
		return (d:p,x)

	c2i (Axiom r name tel t : p) k =
	    do	name' <- freshName DefName name
		tel'  <- telescopeToInternalM tel
		t'    <- bindTelescope tel' $ termToInternalM t
		let d = I.NewAxiom r name' (length tel) (I.telePi tel' t')
		(p, x) <- bindName DefName name' (c2i p k)
		return (d:p, x)

	c2i (Rewrite r es : p) k =
	    do	d <- I.Rewrite r <$> mapM termToInternalM es
		(p, x) <- c2i p k
		return (d:p, x)

	c2i (Mutual r ds : p) k =
	    type2is ts   $ \ds ->
	    val2is ds vs $ \ds ->
	    do	(p,x) <- c2i p k
		return (I.Mutual r ds : p, x)
	    where
		(ts,vs) = unzip ds

	c2i (Abstract r ds : p) k =
	    type2is ts   $ \ds ->
	    val2is ds vs $ \ds ->
	    do	(p,x) <- c2i p k
		return (I.AbstractDefs r ds : p, x)
	    where
		(ts,vs) = unzip ds

	c2i (Def _ dt dv : p) k =
	    type2i dt $ \d ->
		val2i d dv $ \d ->
		do  (p,x) <- c2i p k
		    return (d:p, x)

	val2is :: [I.Definition] -> [DefValue] -> ([I.Definition] -> ConvM a) -> ConvM a
	val2is [] [] k = k []
	val2is (d:ds) (v:vs) k =
	    val2i  d  v  $ \d  ->
	    val2is ds vs $ \ds -> k (d:ds)
	val2is _ _ _ = error "impossible: Syntax.ConcreteToInternal.val2is"

	val2i :: I.Definition -> DefValue -> (I.Definition -> ConvM a) -> ConvM a
	val2i (I.NewData s r x htel tel t _) (DataVal cs) k =
	    do	cs' <- bindTelescope (htel ++ tel)
			(constructorsToInternalM cs)
		bindNames ConName (Set.fromList $ map I.constructorName cs')
		    $ k $ I.NewData s r x htel tel t cs'
	val2i (I.NewSig r x htel tel t ps _) (SigVal cs) k =
	    do	fs  <- Set.unions <$> mapM getSignature ps
		cs' <- bindTelescope (htel ++ tel)
		       $ bindNames VarName fs
		       $ fieldsToInternalM cs
		let fs' = Set.union fs $ Set.fromList (map fieldName cs')
		checkForFieldNameClashes r fs'
		addSignature x fs' $
		    k $ I.NewSig r x htel tel t ps cs'
	val2i (I.NormalDef r x h ar t _) (FunVal cs) k =
	    do	cs <- mapM clauseToInternalM cs
		k $ I.NormalDef r x h ar t cs
	val2i _ _ _ = error "impossible: Syntax.ConcreteToInternal.val2i"

	fieldName (I.Constructor _ x _ _) = x

	type2is :: [DefType] -> ([I.Definition] -> ConvM a) -> ConvM a
	type2is [] k = k []
	type2is (t:ts) k =
	    type2i t   $ \d ->
	    type2is ts $ \ds -> k (d:ds)

	type2i (DataType sized r name htel tel t) k =
	    do	name' <- freshName DataName name
		htel' <- telescopeToInternalM htel
		tel'  <- bindTelescope htel' $ telescopeToInternalM tel
		t'    <- bindTelescope (htel' ++ tel') (termToInternalM t)
		let d = I.NewData sized r name' htel' tel' t' undefined
		bindName DataName name' $ k d
	type2i (SigType r name htel tel t ps) k =
	    do	name' <- freshName DataName name
		htel' <- telescopeToInternalM htel
		tel'  <- bindTelescope htel' $ telescopeToInternalM tel
		t'    <- bindTelescope (htel' ++ tel') (termToInternalM t)
		ps'   <- mapM (getNameOfKind SigName) ps
		let d = I.NewSig r name' htel' tel' t' ps' undefined
		bindName SigName name' $ k d
	type2i (FunType r x a tel t) k =
	    do	let h = length tel
		x'   <- freshName DefName x
		tel' <- telescopeToInternalM tel
		t'   <- bindTelescope tel' $ termToInternalM t
		let d = I.NormalDef r x' h (h + a) (I.telePi tel' t') undefined
		bindName DefName x' $ k d

telescopeToInternalM :: C.Telescope -> ConvM I.Telescope
telescopeToInternalM []		    = return []
telescopeToInternalM ((x,t):tel)    =
    do	t'	<- termToInternalM t
	x'	<- freshName VarName x
	tel'	<- bindName VarName x' $ telescopeToInternalM tel
	return $ (x',t') : tel'

fieldsToInternalM :: [C.Constructor] -> ConvM [I.Constructor]
fieldsToInternalM [] = return []
fieldsToInternalM (c:cs) =
    do	c' <- fieldToInternalM c
	cs' <- bindName VarName (I.constructorName c') $ fieldsToInternalM cs
	return $ c':cs'

fieldToInternalM :: C.Constructor -> ConvM I.Constructor
fieldToInternalM (C.Constructor r x tel t) =
    do	x' <- freshFieldName x
	tel' <- telescopeToInternalM tel
	t'   <- bindTelescope tel' $ termToInternalM t
	return $ I.Constructor r x' tel' t'

constructorsToInternalM :: [C.Constructor] -> ConvM [I.Constructor]
constructorsToInternalM [] = return []
constructorsToInternalM (c:cs) =
    do	c' <- constructorToInternalM c
	cs' <- bindName ConName (I.constructorName c') $ constructorsToInternalM cs
	return $ c':cs'

constructorToInternalM :: C.Constructor -> ConvM I.Constructor
constructorToInternalM (C.Constructor r x tel t) =
    do	x' <- freshName ConName x
	tel' <- telescopeToInternalM tel
	t'   <- bindTelescope tel' $ termToInternalM t
	return $ I.Constructor r x' tel' t'

clauseToInternalM :: Clause -> ConvM I.Clause
clauseToInternalM (Clause _ x c) = clauseToInternalM' x c

clauseToInternalM' :: C.Name -> C.Clause -> ConvM I.Clause
clauseToInternalM' x (C.Clause r hs ps w) =
    do	x' <- getNameOfKind DefName x
	patsToInternalM (hs ++ ps) $ \hps' ->
	    do	w' <- withToInternalM w
		return $ I.Clause r x' hps' w'

withToInternalM :: C.WithClause -> ConvM I.WithClause
withToInternalM (C.NoWith e) = I.NoWith <$> termToInternalM e
withToInternalM (C.With es cs) =
    I.With <$> mapM termToInternalM es
	   <*> mapM (clauseToInternalM' C.DontCare) cs

class HasVars a where
    freeVars :: a -> Set.Set I.Name

instance HasVars a => HasVars [a] where
    freeVars = Set.unions . map freeVars

instance HasVars I.Pattern where
    freeVars (I.WildP _)	    = Set.empty
    freeVars (I.VarP x)		    = Set.singleton x
    freeVars (I.ConP _ _ hs ps)	    = freeVars (hs ++ ps)
    freeVars (I.PairP _ p q)	    = freeVars [p,q]

instance HasVars I.Term where
    freeVars e =
	case e of
	    I.App e1 e2	    -> freeVars [e1,e2]
	    I.HApp e1 e2    -> freeVars [e1,e2]
	    I.Var _ x	    -> Set.singleton x
	    I.Pair _ e1 e2  -> freeVars [e1,e2]
	    I.Project _ e _ -> freeVars e
	    _		    -> Set.empty -- TODO: a bit restrictive...

patsToInternalM :: [C.Pattern] -> ([I.Term] -> ConvM a) -> ConvM a
patsToInternalM [] ret = ret []
patsToInternalM (p:ps) ret =
    patToInternalM p   $ \p'  ->
    patsToInternalM ps $ \ps' ->
    ret (p':ps')

patToInternalM :: C.Pattern -> (I.Term -> ConvM a) -> ConvM a
patToInternalM p ret =
    case p of
	C.WildP r   -> ret $ I.Meta r dontCare
	C.VarP x    ->
	    do	bx@(BName k x') <- freshPatternName x
		bindName k x' $ ret $ bnameToIName bx
	C.ConP r x hs ps ->
	    do	x' <- getName x
		patsToInternalM hs $ \hs' ->
		    patsToInternalM ps $ \ps' ->
		    ret $ x' `happs` hs' `apps` ps'
	C.PairP r p q	->
	    patToInternalM p $ \p' ->
	    patToInternalM q $ \q' ->
	    ret $ I.Pair r p' q'
	p@(C.InfixConP _ _ _) ->
	    do	C.InfixConP p op q <- rotate p
		op' <- getName op
		patToInternalM p $ \p' ->
		    patToInternalM q $ \q' ->
		    ret $ op' `apps` [p',q']
    where
	apps = foldl I.App
	happs = foldl I.HApp

lamPatToInternalM (C.WildP r) ret = ret $ I.WildP r
lamPatToInternalM (C.VarP x) ret =
    do	x' <- freshName VarName x
	bindName VarName x' $ ret $ I.VarP x'
lamPatToInternalM (C.PairP r p q) ret =
    lamPatToInternalM p $ \p' ->
    lamPatToInternalM q $ \q' ->
    ret $ I.PairP r p' q'
lamPatToInternalM p ret =
    fail $ show (getRange p) ++ ": simple pattern required (variable or tuple)"

projectionToInternal :: C.Projection -> I.Projection
projectionToInternal (C.Proj r n)	    = I.Proj r n
projectionToInternal (C.Field (C.Name r x)) = I.Field r x
projectionToInternal _			    = error "impossible: Syntax.ConcreteToInternal.projectionToInternal (Field DontCare)"

termToInternal :: C.Term -> Either String I.Term
termToInternal t = runConvM (termToInternalM t)

-- Note: Maybe we should handle variable shadowing here?
termToInternalM :: C.Term -> ConvM I.Term
termToInternalM = c2i
    where
	c2i (C.App e1 e2)	    = I.App <$> c2i e1 <*> c2i e2
	c2i (C.HApp e1 e2)	    = I.HApp <$> c2i e1 <*> c2i e2
	c2i (C.Id x)		    = getName x
	c2i (C.Lam r p e)	    =
	    lamPatToInternalM p $ \p' ->
	    I.Lam r p' <$> c2i e
	c2i (C.Hash r (C.Univ n))   = return $ I.Hash r (I.intToUniverse n)
	c2i (C.Set r)		    = return $ I.Set r
	c2i (C.ISet r)		    = return $ I.ISet r
	c2i (C.Prop r)		    = return $ I.Prop r
	c2i (C.Type r)		    = return $ I.Type r
	c2i (C.Pi r p a b)	    =
	    do	a' <- c2i a
		lamPatToInternalM p $ \p' ->
		    I.Pi r p' a' <$> c2i b
	c2i (C.Sigma r x a b)   =
	    do	x' <- freshName VarName x
		I.Sigma r x' <$> c2i a
			     <*> bindName VarName x' (c2i b)
	c2i (C.Pair r t u)	    = I.Pair r	<$> c2i t <*> c2i u
	c2i (C.Struct r bs)	    = I.Struct r <$> bindsToInternal bs
	    where
		bindsToInternal [] = return []
		bindsToInternal ((x,hps,ps,e):bs) =
		    do	
			x'  <- freshFieldName x
			(hps', ps', e') <-
			    thread lamPatToInternalM hps $ \hps' ->
			    thread lamPatToInternalM ps $ \ps' ->
			    do	e' <- c2i e
				return (hps', ps', e')
			bs' <- bindName VarName x' $ bindsToInternal bs
			return $ (x',hps',ps',e') : bs'

	c2i (C.Project r t p)	    = I.Project r
					    <$> c2i t
					    <*> return (projectionToInternal p)
        c2i (C.CallPlugin r n es)   =
	    do	scope <- ask
		I.CallPlugin r scope n <$> mapM c2i es
	c2i (C.Meta r)		    = return $ I.Meta r dontCare
	c2i (C.RewriteIn r es e) =
	    I.RewriteIn r <$> mapM termToInternalM es <*> termToInternalM e

	c2i (C.Let r p e)	    =
	    addFixities fixs $ \clashes ->
		if null clashes
		then letToInternal r ds e
		else fail $ "Parse error at " ++ show r ++ ": Repeated fixity declarations for " ++ concat (List.intersperse ", " clashes)
	    where
		(fixs, ds) = definitions p

	c2i (C.Paren _ e)	    = c2i e

	c2i e@(C.InfixApp _ _ _)   = c2i =<< (unInfix <$> rotate e)

	unInfix (C.InfixApp e1 op e2) =
	    C.Id op `C.App` unInfix e1 `C.App` unInfix e2
	unInfix e = e

	letToInternal :: Range -> [Definition] -> C.Term -> ConvM I.Term
	letToInternal r ds e =
	    do	(ds', e') <- definitionsToInternalM ds $ c2i e
		return $ I.Let r ds' e'

data InfixExpr e = InfixApp e C.Name e
		 | Other e

class HasInfixApp e where
    infixView :: e -> InfixExpr e
    unfixView :: InfixExpr e -> e

instance HasInfixApp C.Term where
    infixView (C.InfixApp e1 op e2) = InfixApp e1 op e2
    infixView e			    = Other e

    unfixView (InfixApp e1 op e2)   = C.InfixApp e1 op e2
    unfixView (Other e)		    = e

instance HasInfixApp C.Pattern where
    infixView (C.InfixConP p op q)  = InfixApp p op q
    infixView p			    = Other p

    unfixView (InfixApp p op q)	    = C.InfixConP p op q
    unfixView (Other p)		    = p

rotate :: (HasRange e, HasInfixApp e) => e -> ConvM e
rotate e = case infixView e of
    InfixApp e1 op e2 ->
	do  e1' <- rotate e1
	    case infixView e1' of
		InfixApp e1a op1 e1b ->
		    do  fx  <- getFixity op
			fx1 <- getFixity op1
			rot <- shouldRotate (getRange e) op1 op fx1 fx
			if rot
			    then do e3 <- rotate (unfixView $ InfixApp e1b op e2)
				    return $ unfixView $ InfixApp e1a op1 e3
			    else return $ unfixView $ InfixApp e1' op e2
		_ -> return e
    _	-> return e

shouldRotate r op1 op2 (a1,p1) (a2,p2) =
    case (a1,a2) of
	(C.RightAssoc, C.RightAssoc)	-> return $ p2 >= p1
	(C.LeftAssoc,	C.LeftAssoc)	-> return $ p2 > p1
	_   | p1 == p2			->
		fail $ "Parse error at "
		     ++ show r
		     ++ ":\n  Cannot mix infix applications of (" ++ show op1
		     ++ ") and (" ++ show op2 ++ ")"
	    | otherwise			-> return $ p2 > p1

